class Middleware:
    def __init__(self, request, middlewares=None) -> None:
        self.middlewares = [] if middlewares is None else middlewares
        self.request = request

    def next(self):
        for middleware in self.middlewares:
            handle = middleware(self.request)
            if handle is True:
                continue
            return handle
        return True



import datetime
import functools
import threading
from traceback import format_exception

from renus.core.cache import Cache
from renus.core.log import Log


class Schedule:

    def __init__(self) -> None:
        self._interval = 1
        self._max_try=3
        self._unit = None
        self._at_day = None
        self._at_hour = None
        self._at_minute = None
        self._at_secound = None
        self._forget_day = None
        self._forget_hour = None
        self._forget_minute = None
        self._forget_secound = None
        self._job_list = {}
        self._last_run = Cache().get('schedule_last_run',{})

    def every(self, n=1):
        self._interval = n
        self._max_try = 3
        self._unit = None
        self._at_day = None
        self._at_hour = None
        self._at_minute = None
        self._at_secound = None
        self._forget_day = None
        self._forget_hour = None
        self._forget_minute = None
        self._forget_secound = None
        return self

    def weeks(self):
        self._unit = 'weeks'
        return self

    def days(self):
        self._unit = 'days'
        return self

    def hours(self):
        self._unit = 'hours'
        return self

    def minutes(self):
        self._unit = 'minutes'
        return self

    def secounds(self):
        self._unit = 'secounds'
        return self

    def max_try(self,n:int):
        self._max_try = n
        return self

    def at(self, day:str=None, hour:int=None, minute:int=None, secound:int=None):
        self._at_day = day
        self._at_hour = hour
        self._at_minute = minute
        self._at_secound = secound
        return self

    def forget(self, days:list=None, hours:list=None, minutes:list=None, secounds:list=None):
        self._forget_day = days
        self._forget_hour = hours
        self._forget_minute = minutes
        self._forget_secound = secounds
        return self

    def do(self, key: str, func, *args, **kwargs):
        now = datetime.datetime.utcnow() + datetime.timedelta(hours=3, minutes=30)

        job_func = functools.partial(func, *args, **kwargs)
        try:
            functools.update_wrapper(job_func, func)
        except AttributeError:
            pass

        if self._unit == 'weeks':
            self.handle_weeks(key, job_func, now)
        elif self._unit == 'days':
            self.handle_days(key, job_func, now)
        elif self._unit == 'hours':
            self.handle_hours(key, job_func, now)
        elif self._unit == 'minutes':
            self.handle_minutes(key, job_func, now)
        elif self._unit == 'secounds':
            self.handle_secounds(key, job_func, now)

    def handle_weeks(self, key, job_func, now):
        d = now.strftime('%A')
        h = int(now.strftime('%H'))
        m = int(now.strftime('%M'))
        s = int(now.strftime('%S'))
        if (key in self._last_run and self._last_run[key] < now - datetime.timedelta(
                weeks=self._interval)) or key not in self._last_run:
            if self._at_day is not None and self._at_day != d:
                return None

            if self._at_hour is not None and self._at_hour != h:
                return None

            if self._at_minute is not None and self._at_minute != m:
                return None

            if self._at_secound is not None and self._at_secound != s:
                return None

            self._job_list[key] = job_func
    def handle_days(self, key, job_func, now):
        d = now.strftime('%A')
        h = int(now.strftime('%H'))
        m = int(now.strftime('%M'))
        s = int(now.strftime('%S'))
        if (key in self._last_run and self._last_run[key] < now - datetime.timedelta(
                days=self._interval)) or key not in self._last_run:
            if self._at_hour is not None and self._at_hour != h:
                return None

            if self._at_minute is not None and self._at_minute != m:
                return None

            if self._at_secound is not None and self._at_secound != s:
                return None

            if self._forget_day is not None and d in self._forget_day:
                return None

            self._job_list[key] = job_func

    def handle_hours(self, key, job_func, now):
        h = int(now.strftime('%H'))
        m = int(now.strftime('%M'))
        s = int(now.strftime('%S'))
        if (key in self._last_run and self._last_run[key] < now - datetime.timedelta(
                hours=self._interval)) or key not in self._last_run:

            if self._at_minute is not None and self._at_minute != m:
                return None

            if self._at_secound is not None and self._at_secound != s:
                return None

            if self._forget_hour is not None and h in self._forget_hour:
                return None

            self._job_list[key] = job_func

    def handle_minutes(self, key, job_func, now):
        m = int(now.strftime('%M'))
        s = int(now.strftime('%S'))
        if (key in self._last_run and self._last_run[key] < now - datetime.timedelta(
                minutes=self._interval)) or key not in self._last_run:

            if self._at_secound is not None and self._at_secound != s:
                return None

            if self._forget_minute is not None and m in self._forget_minute:
                return None

            self._job_list[key] = job_func

    def handle_secounds(self, key, job_func, now):
        s = int(now.strftime('%S'))
        if (key in self._last_run and self._last_run[key] < now - datetime.timedelta(
                seconds=self._interval)) or key not in self._last_run:

            if self._forget_secound is not None and s in self._forget_secound:
                return None

            self._job_list[key] = job_func

    def job(self,key, func,logged=False,multi=True,n=0):
        now = datetime.datetime.utcnow() + datetime.timedelta(hours=3, minutes=30)
        self._last_run[key] = now
        Cache().put('schedule_last_run',self._last_run,24*60*60)
        try:
            if multi:
                run_threaded(func)
            else:
                func()
            if logged:
                Log().info(f'schedule job {key} at: {str(now)}')
            self._job_list.pop(key)
        except Exception as exc:
            if n < self._max_try:
                self.job(key,func,logged,multi,n=n+1)
            else:
                Log().error(f'schedule failed [{key}] at: {str(now)} - {str(exc)}')
                self._job_list.pop(key)

    def run(self,logged=False, multi=True):
        jobs = self._job_list.copy()
        for key, func in jobs.items():
            self.job(key,func,logged,multi)

def thread_exception_handler(exc):
   Log().error(f'schedule failed at: '
                f'{str(exc[1])} - '
                f'{format_exception(exc[0],exc[1],exc[2])}'
               )

def run_threaded(job_func):
    threading.excepthook = thread_exception_handler
    t = threading.Thread(target=job_func)
    t.start()




import os, sys

from renus.commands.help import Help

def run():
    arg = sys.argv
    os.system('cls')
    if len(arg) < 2:
        Help()
    if len(arg) == 2:
        getattr(__import__(f"renus.commands.{arg[1]}.run", fromlist=['']), 'run')()
    elif len(arg) > 2:
        getattr(__import__(f"renus.commands.{arg[1]}.run", fromlist=['']), 'run')(arg[2:])


import os
import re


class Default:
    def __init__(self, data):
        self.d = data

    def __call__(self, input):
        if input is None:
            return self.d
        return True


class Convert:
    def __init__(self, func=None, inputs: list = None):
        self.func = func
        self.inputs = inputs

    def __call__(self, input):
        if self.inputs is not None:
            for i in self.inputs:
                if i[0] == input:
                    return i[1]
            return input
        return self.func(input)


def Required(input) -> any:
    if input is not None:
        return True
    return 'required_error'


def String(input) -> any:
    if input is None:
        return True
    t = type(input)
    if t is str:
        return True
    return ['type_error', ['string', str(t)]]


def Bool(input) -> any:
    if input is None:
        return True
    t = type(input)
    if t is bool:
        return True
    return ['type_error', ['bool', str(t)]]


def List(input) -> any:
    if input is None:
        return True
    t = type(input)
    if t is list:
        return True
    return ['type_error', ['list', str(t)]]


def Dict(input) -> any:
    if input is None:
        return True
    t = type(input)
    if t is dict:
        return True
    return ['type_error', ['dict', str(t)]]


def Int(input) -> any:
    if input is None:
        return True
    t = type(input)
    if t is int:
        return True
    return ['type_error', ['int', str(t)]]


def Float(input) -> any:
    if input is None:
        return True
    t = type(input)
    if t is float or t is int:
        return True

    return ['type_error', ['float', str(t)]]


class Len:
    def __init__(self, length: int):
        self.l = length

    def __call__(self, input):
        if input is None:
            return True
        length = len(input)
        if length == self.l:
            return True
        return ['len_error', [self.l, length]]


class MinLen:
    def __init__(self, length: int):
        self.l = length

    def __call__(self, input):
        if input is None:
            return True
        length = len(input)
        if length >= self.l:
            return True
        return ['min_len_error', [self.l, length]]


class MaxLen:
    def __init__(self, length: int):
        self.l = length

    def __call__(self, input):
        if input is None:
            return True
        length = len(input)
        if length <= self.l:
            return True
        return ['max_len_error', [self.l, length]]


class Eq:
    def __init__(self, length: float):
        self.l = length

    def __call__(self, input):
        if input is None:
            return True

        if input == self.l:
            return True
        return ['eq_error', [self.l, input]]


class Min:
    def __init__(self, length: float):
        self.l = length

    def __call__(self, input):
        if input is None:
            return True

        if input >= self.l:
            return True
        return ['min_error', [self.l, input]]


class Max:
    def __init__(self, length: float):
        self.l = length

    def __call__(self, input):
        if input is None:
            return True

        if input <= self.l:
            return True
        return ['max_error', [self.l, input]]


def Numeric(input) -> any:
    if input is None:
        return True
    if re.compile('[\d]+').fullmatch(str(input)) != None:
        return True
    return ['numeric_error', [input]]


def Accepted(input) -> any:
    if input is None:
        return True
    if String(input) == True:
        input = input.lower()
    acceptable = ['yes', 'on', '1', 1, True, 'true']
    if input in acceptable:
        return True
    return ['accepted_error', [input]]


class Email:
    def __init__(self, allow_domain: list = None):
        self.allow_domain = [] if allow_domain is None else allow_domain

    def __call__(self, input):

        if input is None:
            return True

        is_string = String(input)
        if is_string == True:
            input = input.lower()
        else:
            return is_string

        parse = input.split('@')
        if len(parse) < 2:
            return ['email_at_error', [input]]
        name, domain = parse
        parse_domain = domain.split('.')
        if len(parse_domain) < 2:
            return ['email_domain_name_error', [domain]]
        if len(parse_domain[1]) < 2:
            return ['email_domain_name_error', [domain]]
        if len(self.allow_domain) > 0:
            if domain not in self.allow_domain:
                return ['email_domain_allow_error', [domain, str(self.allow_domain)]]
        if len(name) < 3:
            return ['email_name_error', [3, input]]

        return True


class In:
    def __init__(self, *vars):
        self.vars = list(vars)

    def __call__(self, input):
        if input is None:
            return True
        if input in self.vars:
            return True
        return ['in_error', [input, str(self.vars)]]


class NotIn:
    def __init__(self, *vars):
        self.vars = list(vars)

    def __call__(self, input):
        if input is None:
            return True

        if input not in self.vars:
            return True
        return ['not_in_error', [input, str(self.vars)]]


class Regex:
    def __init__(self, rule: str, msg: str = None):
        self.rule = rule
        self.msg = msg

    def __call__(self, input):
        if input is None:
            return True
        msg = None

        if re.compile(self.rule).fullmatch(input) is not None:
            return True
        return ['regex_error', [input, msg]]


def Url(input):
    if input is None:
        return True
    string = String(input)
    if string is not True:
        return string
    pattern = r"(?i)\b((?:https?://|www\d{0,3}[.]|[a-z0-9.\-]+[.][a-z]{2,4}/)(?:[^\s()<>]+|\(([^\s()<>]+|(\([^\s()<>]+\)))*\))+(?:\(([^\s()<>]+|(\([^\s()<>]+\)))*\)|[^\s`!()\[\]{};:'\".,<>?«»“”‘’]))"

    if re.compile(pattern).fullmatch(input) is not None:
        return True
    return ['url_error', [input]]


def Ip(input):
    if input is None:
        return True
    string = String(input)
    if string is not True:
        return string
    ipv = re.match(r"^(\d{1,3})\.(\d{1,3})\.(\d{1,3})\.(\d{1,3})$", input)
    if bool(ipv) and all(map(lambda n: 0 <= int(n) <= 255, ipv.groups())):
        return True

    return ['ip_error', [input]]


class Extension:
    def __init__(self, extensions: list = None):
        self.allowed = ['jpeg', 'jpg', 'png', 'gif', 'svg', 'webp'] if extensions is None else extensions

    def __call__(self, input):
        if input is None:
            return True
        string = String(input)
        if string is True:
            return self.__check(input)

        for img in input:
            c = self.__check(img)
            if c is not True:
                return c
        return True

    def __check(self, string):
        i = string.split('.')
        if i[-1] not in self.allowed:
            return ['image_error', [string]]
        return True


class Unique:
    def __init__(self, model, field, current_id=None):
        self.model = model
        self.field = field
        self.current_id = current_id

    def __call__(self, input):

        if input is None:
            return True

        model = self.model.where({
            self.field: input
        }).select('_id').first()
        if model is None:
            return True

        if self.current_id is not None and str(self.current_id) == str(model['_id']):
            return True
        return ['unique_error', [input]]


class Exists:
    def __init__(self, model, field):
        self.model = model
        self.field = field

    def __call__(self, input):

        if input is None:
            return True

        model = self.model.where({
            self.field: input
        }).select('_id').first()
        if model is not None:
            return True
        return ['exists_error', [input]]


class ExistsCount:
    def __init__(self, model, field, count: int = None, minCount: int = None, maxCount: int = None):
        self.model = model
        self.field = field
        self.count = count
        self.minCount = minCount
        self.maxCount = maxCount

    def __call__(self, input):
        if input is None:
            return True

        count = self.model.where({
            self.field: input
        }).select('_id').count()

        if (self.count is not None and self.count == count) or (
                self.minCount is not None and self.minCount <= count) or (
                self.maxCount is not None and self.maxCount >= count):
            return True
        return ['exists_count_error', [input]]


class FileSize:
    def __init__(self, max_byte: int = 0, min_byte: int = 0, delete=True, replace: list = None, item_dict: str = None):
        """
        @param max_byte: maximum allowed Bytes
        @param min_byte:minimum allowed Bytes
        @param delete: delete file if not passed
        @param replace: replace file path. default is [('storage/img/','storage/'),('storage/','storage/public/')]
        @param item_dict: if file path in a dict. ex {url:'',meta:''} => item_dict='url'
        """
        if replace is None:
            replace = [('storage/img/', 'storage/'), ('storage/', 'storage/public/')]
        self.delete = delete
        self.maxByte = max_byte
        self.minByte = min_byte
        self.replace = replace
        self.item_dict = item_dict

    def __call__(self, input):
        if input is None:
            return True
        t = type(input)
        if t is str:
            input = [input]
        passed = True
        if self.maxByte:
            for item in input:
                tt = type(item)
                if tt is dict:
                    link = item[self.item_dict]
                else:
                    link = item
                for rep in self.replace:
                    link = link.replace(rep[0], rep[1])
                state = os.stat(link)
                if state.st_size <= self.maxByte:
                    continue
                passed = ['max_file_size_error', [item, state.st_size, self.maxByte]]
        if self.minByte:
            for item in input:
                tt = type(item)
                if tt is dict:
                    link = item[self.item_dict]
                else:
                    link = item
                for rep in self.replace:
                    link = link.replace(rep[0], rep[1])
                state = os.stat(link)
                if state.st_size >= self.minByte:
                    continue
                passed = ['min_file_size_error', [item, state.st_size, self.minByte]]
        if passed is not True and self.delete:
            for item in input:
                tt = type(item)
                if tt is dict:
                    link = item[self.item_dict]
                else:
                    link = item
                for rep in self.replace:
                    link = link.replace(rep[0], rep[1])

                os.remove(link)

        return passed



import datetime
import decimal
import json
import uuid

try:
    from bson import ObjectId
except:
    def ObjectId(s):
        raise


class jsonEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, datetime.datetime):
            return o.isoformat()+'Z'
        elif isinstance(o, datetime.date):
            return o.isoformat() + 'T00:00:00Z'
        elif isinstance(o, datetime.time):
            r = o.isoformat()
            if o.microsecond:
                r = r[:12]
            return r

        elif isinstance(o, (dict, list, tuple, str, int, float, bool, type(None))):
            return super().default(o)
        elif hasattr(o, '__dict__'):
            return dict(o)
        else:
            return str(o)


def json_decoder(value):
    if isinstance(value, dict):
        for k, v in value.items():
            value[k] = json_decoder(v)
    elif isinstance(value, list):
        for index, row in enumerate(value):
            value[index] = json_decoder(row)
    elif isinstance(value, str) and value:
        try:
            value = datetime.datetime.fromisoformat(value)
        except:
            try:
                value = ObjectId(value)
            except:
                try:
                    value = decimal.Decimal(value)
                except:
                    try:
                        value = uuid.UUID(value)
                    except:
                        pass

    return value



from setuptools import setup, find_packages

setup(
    name='renus',
    version='1.6.0',
    description='Renus Core Framework',
    url='https://github.com/smkoBa/renus',
    author='Smko Bayazidi',
    author_email='ba.smko@gmail.com',
    license='BSD',
    packages=find_packages(),
    include_package_data=True,
    python_requires=">=3.8",
    install_requires=['anyio', 'python-multipart', 'pymongo', 'requests', 'setuptools']
)
